@@ -2,35 +2,41 @@ use std::collections::{HashMap, HashSet};
 use std::fs::File;
 use std::io;
 use std::io::{Read, Write};
-use std::path::{Path, PathBuf};
+use std::path::Path;
 
 use chrono::NaiveDateTime;
 use zip::write::FileOptions;
 
-use crate::attrmap2::AttrMap2Trait;
 use crate::error::OdsError;
 use crate::format::FormatPartType;
-use crate::io::tmp2zip::{TempWrite, TempZip};
 use crate::io::xmlwriter::XmlWriter;
+use crate::io::zip_out::{ZipOut, ZipWrite};
 use crate::refs::{cellranges_string, CellRange};
+use crate::settings::{ConfigItem, ConfigItemType, ConfigValue};
 use crate::style::{
     CellStyle, ColStyle, FontFaceDecl, GraphicStyle, HeaderFooter, MasterPage, PageStyle,
     ParagraphStyle, RowStyle, StyleOrigin, StyleUse, TableStyle, TextStyle,
 };
 use crate::xmltree::{XmlContent, XmlTag};
-use crate::{ucell, SCell, Sheet, Value, ValueFormat, ValueType, Visibility, WorkBook};
+use crate::{ucell, Length, SCell, Sheet, Value, ValueFormat, ValueType, Visibility, WorkBook};
 
-type OdsWriter = TempZip;
-type XmlOdsWriter<'a> = XmlWriter<TempWrite<'a>>;
+type OdsWriter = ZipOut;
+type XmlOdsWriter<'a> = XmlWriter<ZipWrite<'a>>;
 
 /// Writes the ODS file.
 ///
 /// All the parts are written to a temp directory and then zipped together.
 ///
-pub fn write_ods<P: AsRef<Path>>(book: &WorkBook, ods_path: P) -> Result<(), OdsError> {
-    let mut zip_writer = TempZip::new(ods_path.as_ref())?;
+pub fn write_ods<P: AsRef<Path>>(book: &mut WorkBook, ods_path: P) -> Result<(), OdsError> {
+    store_derived(book)?;
+
+    let mut zip_writer = ZipOut::new(ods_path.as_ref())?;
 
     let mut file_set = HashSet::<String>::new();
+    // never copy these
+    file_set.insert(String::from("settings.xml"));
+    file_set.insert(String::from("styles.xml"));
+    file_set.insert(String::from("content.xml"));
 
     if let Some(orig) = &book.file {
         copy_workbook(&orig, &mut file_set, &mut zip_writer)?;
@@ -40,18 +46,107 @@ pub fn write_ods<P: AsRef<Path>>(book: &WorkBook, ods_path: P) -> Result<(), Ods
     write_manifest(&book, &mut zip_writer, &mut file_set)?;
     write_manifest_rdf(&book, &mut zip_writer, &mut file_set)?;
     write_meta(&book, &mut zip_writer, &mut file_set)?;
-    //write_settings(&mut zip_writer, &mut file_set)?;
-    //write_configurations(&mut zip_writer, &mut file_set)?;
-    write_ods_styles(&book, &mut zip_writer, &mut file_set)?;
-    write_ods_content(&book, &mut zip_writer, &mut file_set)?;
+    write_settings(&book, &mut zip_writer)?;
+    // not in use any more, just ignore
+    // write_configurations(&mut zip_writer, &mut file_set)?;
+    write_ods_styles(&book, &mut zip_writer)?;
+    write_ods_content(&book, &mut zip_writer)?;
 
     zip_writer.zip()?;
 
     Ok(())
 }
 
+#[allow(clippy::collapsible_else_if)]
+#[allow(clippy::collapsible_if)]
+fn store_derived(book: &mut WorkBook) -> Result<(), OdsError> {
+    let mut config = book.config.detach(0);
+
+    let bc = config.create_path(&[
+        ("ooo:view-settings", ConfigItemType::Set),
+        ("Views", ConfigItemType::Vec),
+        ("0", ConfigItemType::Entry),
+    ]);
+    bc.insert("ActiveTable", book.config().active_table.clone());
+    bc.insert("HasSheetTabs", book.config().has_sheet_tabs);
+    bc.insert("ShowGrid", book.config().show_grid);
+    bc.insert("ShowPageBreaks", book.config().show_page_breaks);
+
+    for i in 0..book.num_sheets() {
+        let mut sheet = book.detach_sheet(i);
+
+        // Set the column widths.
+        for ch in sheet.col_header.values_mut() {
+            // Any non default values?
+            if ch.width() != Length::Default {
+                if ch.style().is_none() {
+                    let colstyle = book.add_colstyle(ColStyle::empty());
+                    ch.set_style(&colstyle);
+                }
+            }
+
+            // Write back to the style.
+            if let Some(style_name) = ch.style() {
+                if let Some(style) = book.colstyle_mut(style_name) {
+                    if ch.width() == Length::Default {
+                        style.set_use_optimal_col_width(true);
+                        style.set_col_width(Length::Default);
+                    } else {
+                        style.set_col_width(ch.width());
+                    }
+                }
+            }
+        }
+
+        for rh in sheet.row_header.values_mut() {
+            if rh.height() != Length::Default {
+                if rh.style().is_none() {
+                    let rowstyle = book.add_rowstyle(RowStyle::empty());
+                    rh.set_style(&rowstyle);
+                }
+            }
+
+            if let Some(style_name) = rh.style() {
+                if let Some(style) = book.rowstyle_mut(style_name) {
+                    if rh.height() == Length::Default {
+                        style.set_use_optimal_row_height(true);
+                        style.set_row_height(Length::Default);
+                    } else {
+                        style.set_row_height(rh.height());
+                    }
+                }
+            }
+        }
+
+        let bc = config.create_path(&[
+            ("ooo:view-settings", ConfigItemType::Set),
+            ("Views", ConfigItemType::Vec),
+            ("0", ConfigItemType::Entry),
+            ("Tables", ConfigItemType::Map),
+            (sheet.name().as_str(), ConfigItemType::Entry),
+        ]);
+
+        bc.insert("CursorPositionX", sheet.config().cursor_x);
+        bc.insert("CursorPositionY", sheet.config().cursor_y);
+        bc.insert("HorizontalSplitMode", sheet.config().hor_split_mode as i16);
+        bc.insert("VerticalSplitMode", sheet.config().vert_split_mode as i16);
+        bc.insert("HorizontalSplitPosition", sheet.config().hor_split_pos);
+        bc.insert("VerticalSplitPosition", sheet.config().vert_split_pos);
+        bc.insert("ActiveSplitRange", sheet.config().active_split_range);
+        bc.insert("ZoomType", sheet.config().zoom_type);
+        bc.insert("ZoomValue", sheet.config().zoom_value);
+        bc.insert("ShowGrid", sheet.config().show_grid);
+
+        book.attach_sheet(sheet);
+    }
+
+    book.config.attach(config);
+
+    Ok(())
+}
+
 fn copy_workbook(
-    ods_orig_name: &PathBuf,
+    ods_orig_name: &Path,
     file_set: &mut HashSet<String>,
     zip_writer: &mut OdsWriter,
 ) -> Result<(), OdsError> {
@@ -276,32 +371,217 @@ fn write_meta(
     Ok(())
 }
 
-//fn write_settings(zip_out: &mut ZipWriter<BufWriter<File>>, file_set: &mut HashSet<String>) -> Result<(), OdsError> {
-//    if !file_set.contains("settings.xml") {
-//        file_set.insert(String::from("settings.xml"));
-//
-//        zip_out.start_file("settings.xml", FileOptions::default())?;
-//
-//        let mut xml_out = XmlWriter::new(zip_out);
-//
-//        xml_out.write_event(Event::Decl(BytesDecl::new(b"1.0", Some(b"UTF-8"), None)))?;
-//        xml_out.write(b"\n")?;
-//
-//        xml_out.write_event(xml_start_a("office:document-settings", vec![
-//            ("xmlns:office", String::from("urn:oasis:names:tc:opendocument:xmlns:office:1.0")),
-//            ("xmlns:ooo", String::from("http://openoffice.org/2004/office")),
-//            ("xmlns:config", String::from("urn:oasis:names:tc:opendocument:xmlns:config:1.0")),
-//            ("office:version", String::from("1.2")),
-//        ]))?;
-//
-//        xml_out.write_event(xml_start("office:settings"))?;
-//        xml_out.write_event(xml::xml_end("office:settings"))?;
-//
-//        xml_out.write_event(xml::xml_end("office:document-settings"))?;
-//    }
-//
-//    Ok(())
-//}
+fn write_settings(book: &WorkBook, zip_out: &mut OdsWriter) -> Result<(), OdsError> {
+    let w = zip_out.start_file("settings.xml", FileOptions::default())?;
+
+    let mut xml_out = XmlWriter::new(w);
+
+    xml_out.dtd("UTF-8")?;
+
+    xml_out.elem("office:document-settings")?;
+    xml_out.attr(
+        "xmlns:office",
+        "urn:oasis:names:tc:opendocument:xmlns:office:1.0",
+    )?;
+    xml_out.attr("xmlns:ooo", "http://openoffice.org/2004/office")?;
+    xml_out.attr(
+        "xmlns:config",
+        "urn:oasis:names:tc:opendocument:xmlns:config:1.0",
+    )?;
+    xml_out.attr("office:version", book.version())?;
+    xml_out.elem("office:settings")?;
+
+    for (name, item) in book.config.iter() {
+        match item {
+            ConfigItem::Value(_) => {
+                panic!("office-settings must not contain config-item");
+            }
+            ConfigItem::Set(_) => write_config_item_set(name, item, &mut xml_out)?,
+            ConfigItem::Vec(_) => {
+                panic!("office-settings must not contain config-item-map-index")
+            }
+            ConfigItem::Map(_) => {
+                panic!("office-settings must not contain config-item-map-named")
+            }
+            ConfigItem::Entry(_) => {
+                panic!("office-settings must not contain config-item-map-entry")
+            }
+        }
+    }
+
+    xml_out.end_elem("office:settings")?;
+    xml_out.end_elem("office:document-settings")?;
+
+    xml_out.close()?;
+
+    Ok(())
+}
+
+fn write_config_item_set(
+    name: &String,
+    set: &ConfigItem,
+    xml_out: &mut XmlOdsWriter,
+) -> Result<(), OdsError> {
+    xml_out.elem("config:config-item-set")?;
+    xml_out.attr("config:name", name)?;
+
+    for (name, item) in set.iter() {
+        match item {
+            ConfigItem::Value(value) => write_config_item(name, &value, xml_out)?,
+            ConfigItem::Set(_) => write_config_item_set(name, &item, xml_out)?,
+            ConfigItem::Vec(_) => write_config_item_map_indexed(name, &item, xml_out)?,
+            ConfigItem::Map(_) => write_config_item_map_named(name, &item, xml_out)?,
+            ConfigItem::Entry(_) => {
+                panic!("config-item-set must not contain config-item-map-entry")
+            }
+        }
+    }
+
+    xml_out.end_elem("config:config-item-set")?;
+
+    Ok(())
+}
+
+fn write_config_item_map_indexed(
+    name: &String,
+    vec: &ConfigItem,
+    xml_out: &mut XmlOdsWriter,
+) -> Result<(), OdsError> {
+    xml_out.elem("config:config-item-map-indexed")?;
+    xml_out.attr("config:name", name)?;
+
+    let mut index = 0;
+    loop {
+        let index_str = index.to_string();
+        if let Some(item) = vec.get(&index_str) {
+            match item {
+                ConfigItem::Value(value) => write_config_item(name, &value, xml_out)?,
+                ConfigItem::Set(_) => {
+                    panic!("config-item-map-index must not contain config-item-set")
+                }
+                ConfigItem::Vec(_) => {
+                    panic!("config-item-map-index must not contain config-item-map-index")
+                }
+                ConfigItem::Map(_) => {
+                    panic!("config-item-map-index must not contain config-item-map-named")
+                }
+                ConfigItem::Entry(_) => write_config_item_map_entry(None, &item, xml_out)?,
+            }
+        } else {
+            break;
+        }
+
+        index += 1;
+    }
+
+    xml_out.end_elem("config:config-item-map-indexed")?;
+
+    Ok(())
+}
+
+fn write_config_item_map_named(
+    name: &String,
+    map: &ConfigItem,
+    xml_out: &mut XmlOdsWriter,
+) -> Result<(), OdsError> {
+    xml_out.elem("config:config-item-map-named")?;
+    xml_out.attr("config:name", name)?;
+
+    for (name, item) in map.iter() {
+        match item {
+            ConfigItem::Value(value) => write_config_item(name, &value, xml_out)?,
+            ConfigItem::Set(_) => {
+                panic!("config-item-map-index must not contain config-item-set")
+            }
+            ConfigItem::Vec(_) => {
+                panic!("config-item-map-index must not contain config-item-map-index")
+            }
+            ConfigItem::Map(_) => {
+                panic!("config-item-map-index must not contain config-item-map-named")
+            }
+            ConfigItem::Entry(_) => write_config_item_map_entry(Some(name), &item, xml_out)?,
+        }
+    }
+
+    xml_out.end_elem("config:config-item-map-named")?;
+
+    Ok(())
+}
+
+fn write_config_item_map_entry(
+    name: Option<&String>,
+    map_entry: &ConfigItem,
+    xml_out: &mut XmlOdsWriter,
+) -> Result<(), OdsError> {
+    xml_out.elem("config:config-item-map-entry")?;
+    if let Some(name) = name {
+        xml_out.attr("config:name", name)?;
+    }
+
+    for (name, item) in map_entry.iter() {
+        match item {
+            ConfigItem::Value(value) => write_config_item(name, &value, xml_out)?,
+            ConfigItem::Set(_) => write_config_item_set(name, &item, xml_out)?,
+            ConfigItem::Vec(_) => write_config_item_map_indexed(name, &item, xml_out)?,
+            ConfigItem::Map(_) => write_config_item_map_named(name, &item, xml_out)?,
+            ConfigItem::Entry(_) => {
+                panic!("config:config-item-map-entry must not contain config-item-map-entry")
+            }
+        }
+    }
+
+    xml_out.end_elem("config:config-item-map-entry")?;
+
+    Ok(())
+}
+
+fn write_config_item(
+    name: &String,
+    value: &ConfigValue,
+    xml_out: &mut XmlOdsWriter,
+) -> Result<(), OdsError> {
+    xml_out.elem("config:config-item")?;
+    xml_out.attr("config:name", name)?;
+
+    match value {
+        ConfigValue::Base64Binary(v) => {
+            xml_out.attr("config:type", "base64Binary")?;
+            xml_out.text(v)?;
+        }
+        ConfigValue::Boolean(v) => {
+            xml_out.attr("config:type", "boolean")?;
+            xml_out.text(&v.to_string())?;
+        }
+        ConfigValue::DateTime(v) => {
+            xml_out.attr("config:type", "datetime")?;
+            xml_out.text(&v.format("%Y-%m-%dT%H:%M:%S%.f").to_string())?;
+        }
+        ConfigValue::Double(v) => {
+            xml_out.attr("config:type", "double")?;
+            xml_out.text(&v.to_string())?;
+        }
+        ConfigValue::Int(v) => {
+            xml_out.attr("config:type", "int")?;
+            xml_out.text(&v.to_string())?;
+        }
+        ConfigValue::Long(v) => {
+            xml_out.attr("config:type", "long")?;
+            xml_out.text(&v.to_string())?;
+        }
+        ConfigValue::Short(v) => {
+            xml_out.attr("config:type", "short")?;
+            xml_out.text(&v.to_string())?;
+        }
+        ConfigValue::String(v) => {
+            xml_out.attr("config:type", "string")?;
+            xml_out.text(v)?;
+        }
+    }
+
+    xml_out.end_elem("config:config-item")?;
+
+    Ok(())
+}
 
 //fn write_configurations(zip_out: &mut ZipWriter<BufWriter<File>>, file_set: &mut HashSet<String>) -> Result<(), OdsError> {
 //    if !file_set.contains("Configurations2") {
@@ -333,13 +613,7 @@ fn write_meta(
 //    Ok(())
 //}
 
-fn write_ods_styles(
-    book: &WorkBook,
-    zip_out: &mut OdsWriter,
-    file_set: &mut HashSet<String>,
-) -> Result<(), OdsError> {
-    file_set.insert(String::from("styles.xml"));
-
+fn write_ods_styles(book: &WorkBook, zip_out: &mut OdsWriter) -> Result<(), OdsError> {
     let w = zip_out.start_file("styles.xml", FileOptions::default())?;
 
     let mut xml_out = XmlWriter::new(w);
@@ -457,13 +731,7 @@ fn write_ods_styles(
     Ok(())
 }
 
-fn write_ods_content(
-    book: &WorkBook,
-    zip_out: &mut OdsWriter,
-    file_set: &mut HashSet<String>,
-) -> Result<(), OdsError> {
-    file_set.insert(String::from("content.xml"));
-
+fn write_ods_content(book: &WorkBook, zip_out: &mut OdsWriter) -> Result<(), OdsError> {
     let w = zip_out.start_file("content.xml", FileOptions::default())?;
     let mut xml_out = XmlWriter::new(w);
 
@@ -580,7 +848,7 @@ fn write_ods_content(
     }
 
     for sheet in &book.sheets {
-        write_sheet(&book, &sheet, &mut xml_out)?;
+        write_sheet(&book, sheet, &mut xml_out)?;
     }
 
     // extra tags. pass through only
@@ -698,9 +966,6 @@ fn write_sheet(book: &WorkBook, sheet: &Sheet, xml_out: &mut XmlOdsWriter) -> Re
             *cur_col - last_c
         };
 
-        // println!("cell first={} {},{} < *{},{}* < {},{} ", first_cell, last_r, last_c, cur_row, cur_col, next_r, next_c);
-        // println!("     backward {},{} forward {},{}", backward_dr, backward_dc, forward_dr, forward_dc);
-
         // After the first cell there is always an open row tag that
         // needs to be closed.
         if backward_dr > 0 && !first_cell {
@@ -804,6 +1069,9 @@ fn write_start_current_row(
 
     xml_out.elem("table:table-row")?;
     if let Some(row_header) = sheet.row_header.get(&cur_row) {
+        if row_header.repeat > 1 {
+            xml_out.attr_esc("table:number-rows-repeated", &row_header.repeat.to_string())?;
+        }
         if let Some(rowstyle) = row_header.style() {
             xml_out.attr_esc("table:style-name", rowstyle.as_str())?;
         }
@@ -931,7 +1199,7 @@ fn write_empty_row(
     cur_row: ucell,
     empty_count: u32,
     max_cell: (u32, u32),
-    xml_out: &mut XmlWriter<TempWrite>,
+    xml_out: &mut XmlOdsWriter,
 ) -> Result<(), OdsError> {
     xml_out.elem("table:table-row")?;
     xml_out.attr("table:number-rows-repeated", &empty_count.to_string())?;
