@@ -58,7 +58,7 @@
 //!
 
 //!
-//! spreadsheet_ods::write_ods(&wb, "test_out/tryout.ods");
+//! spreadsheet_ods::write_ods(&mut wb, "test_out/tryout.ods");
 //!
 //! ```
 //! This does not cover the entire ODS spec.
@@ -143,6 +143,37 @@
 
 #![doc(html_root_url = "https://docs.rs/spreadsheet-ods/0.4.0")]
 
+use std::collections::{BTreeMap, HashMap};
+use std::convert::TryFrom;
+use std::fmt;
+use std::fmt::{Display, Formatter};
+use std::path::PathBuf;
+use std::str::FromStr;
+
+use chrono::Duration;
+use chrono::{NaiveDate, NaiveDateTime};
+#[cfg(feature = "use_decimal")]
+use rust_decimal::prelude::*;
+#[cfg(feature = "use_decimal")]
+use rust_decimal::Decimal;
+
+use crate::ds::detach::Detach;
+pub use crate::ds::detach::Detached;
+pub use crate::error::OdsError;
+pub use crate::format::{ValueFormat, ValueFormatRef};
+pub use crate::io::{read_ods, write_ods};
+pub use crate::refs::{CellRange, CellRef, ColRange, RowRange};
+use crate::settings::Config;
+pub use crate::style::units::{Angle, Length};
+pub use crate::style::{CellStyle, CellStyleRef};
+use crate::style::{
+    ColStyle, ColStyleRef, FontFaceDecl, GraphicStyle, GraphicStyleRef, MasterPage, MasterPageRef,
+    PageStyle, PageStyleRef, ParagraphStyle, ParagraphStyleRef, RowStyle, RowStyleRef, TableStyle,
+    TableStyleRef, TextStyle, TextStyleRef,
+};
+use crate::text::TextTag;
+use crate::xmltree::XmlTag;
+
 #[macro_use]
 mod attr_macro;
 #[macro_use]
@@ -151,41 +182,17 @@ mod unit_macro;
 mod ref_macro;
 mod attrmap2;
 pub mod defaultstyles;
+mod ds;
 pub mod error;
 pub mod format;
 pub mod formula;
 mod io;
 pub mod refs;
+mod settings;
 pub mod style;
 pub mod text;
 pub mod xmltree;
 
-pub use crate::error::OdsError;
-pub use crate::format::{ValueFormat, ValueFormatRef};
-pub use crate::io::{read_ods, write_ods};
-pub use crate::refs::{CellRange, CellRef, ColRange, RowRange};
-pub use crate::style::units::{Angle, Length};
-pub use crate::style::{CellStyle, CellStyleRef};
-
-use crate::style::{
-    ColStyle, ColStyleRef, FontFaceDecl, GraphicStyle, GraphicStyleRef, MasterPage, MasterPageRef,
-    PageStyle, PageStyleRef, ParagraphStyle, ParagraphStyleRef, RowStyle, RowStyleRef, TableStyle,
-    TableStyleRef, TextStyle, TextStyleRef,
-};
-use crate::text::TextTag;
-use crate::xmltree::XmlTag;
-use chrono::Duration;
-use chrono::{NaiveDate, NaiveDateTime};
-#[cfg(feature = "use_decimal")]
-use rust_decimal::prelude::*;
-#[cfg(feature = "use_decimal")]
-use rust_decimal::Decimal;
-use std::collections::{BTreeMap, HashMap};
-use std::fmt;
-use std::fmt::{Display, Formatter};
-use std::path::PathBuf;
-use std::str::FromStr;
-
 /// Cell index type for row/column indexes.
 #[allow(non_camel_case_types)]
 pub type ucell = u32;
@@ -194,14 +201,17 @@ pub type ucell = u32;
 #[derive(Clone, Default)]
 pub struct WorkBook {
     /// The data.
-    sheets: Vec<Sheet>,
+    sheets: Vec<Detach<Sheet>>,
 
     // ODS Version
     version: String,
 
-    //// FontDecl hold the style:font-face elements
+    /// FontDecl hold the style:font-face elements
     fonts: HashMap<String, FontFaceDecl>,
 
+    /// Auto-Styles. Maps the prefix to a number.
+    autonum: HashMap<String, u32>,
+
     /// Styles hold the style:style elements.
     tablestyles: HashMap<String, TableStyle>,
     rowstyles: HashMap<String, RowStyle>,
@@ -224,6 +234,12 @@ pub struct WorkBook {
     pagestyles: HashMap<String, PageStyle>,
     masterpages: HashMap<String, MasterPage>,
 
+    /// Configuration data. Internal cache for all values.
+    /// Mapped into WorkBookConfig, SheetConfig.
+    config: Detach<Config>,
+    /// User modifiable config.
+    workbook_config: WorkBookConfig,
+
     /// Original file if this book was read from one.
     /// This is used when writing to copy all additional
     /// files except content.xml
@@ -280,12 +296,39 @@ impl fmt::Debug for WorkBook {
     }
 }
 
+/// Autogenerate a stylename. Runs a counter with the prefix and
+/// checks for existence.
+fn auto_style_name<T>(
+    autonum: &mut HashMap<String, u32>,
+    prefix: &str,
+    styles: &HashMap<String, T>,
+) -> String {
+    let mut cnt = if let Some(n) = autonum.get(prefix) {
+        n + 1
+    } else {
+        0
+    };
+
+    let style_name = loop {
+        let style_name = format!("{}{}", prefix, cnt);
+        if !styles.contains_key(&style_name) {
+            break style_name;
+        }
+        cnt += 1;
+    };
+
+    autonum.insert(prefix.to_string(), cnt);
+
+    style_name
+}
+
 impl WorkBook {
     pub fn new() -> Self {
         WorkBook {
             sheets: Default::default(),
             version: "1.3".to_string(),
             fonts: Default::default(),
+            autonum: Default::default(),
             tablestyles: Default::default(),
             rowstyles: Default::default(),
             colstyles: Default::default(),
@@ -297,6 +340,8 @@ impl WorkBook {
             def_styles: Default::default(),
             pagestyles: Default::default(),
             masterpages: Default::default(),
+            config: Default::default(),
+            workbook_config: Default::default(),
             file: None,
             extra: vec![],
         }
@@ -312,42 +357,79 @@ impl WorkBook {
         self.version = version;
     }
 
+    pub fn config(&self) -> &WorkBookConfig {
+        &self.workbook_config
+    }
+
+    pub fn config_mut(&mut self) -> &mut WorkBookConfig {
+        &mut self.workbook_config
+    }
+
     /// Number of sheets.
     pub fn num_sheets(&self) -> usize {
         self.sheets.len()
     }
 
+    /// Detaches a sheet.
+    /// Useful if you have to make mutating calls to the workbook and
+    /// the sheet intermixed.
+    ///
+    /// Warning
+    ///
+    /// The sheet has to be re-attached before saving the workbook.
+    ///
+    /// Panics
+    ///
+    /// Panics if the sheet has already been detached.
+    /// Panics if n is out of bounds.
+    pub fn detach_sheet(&mut self, n: usize) -> Detached<usize, Sheet> {
+        self.sheets[n].detach(n)
+    }
+
+    /// Reattaches the sheet in the place it was before.
+    ///
+    /// Panics
+    ///
+    /// Panics if n is out of bounds.
+    pub fn attach_sheet(&mut self, sheet: Detached<usize, Sheet>) {
+        self.sheets[Detached::key(&sheet)].attach(sheet)
+    }
+
+    /// Returns a certain sheet.
+    ///
+    /// Panics
+    ///
+    /// Panics if n is out of bounds.
+    pub fn sheet(&self, n: usize) -> &Sheet {
+        &self.sheets[n].as_ref()
+    }
+
     /// Returns a certain sheet.
     ///
     /// Panics
     ///
     /// Panics if n does not exist.
-    pub fn sheet(&self, n: usize) -> &Sheet {
-        &self.sheets[n]
-    }
-
-    /// Returns a certain sheet.
-    ///
-    /// Panic
-    ///
-    /// Panics if n does not exist.
     pub fn sheet_mut(&mut self, n: usize) -> &mut Sheet {
-        &mut self.sheets[n]
+        self.sheets[n].as_mut()
     }
 
     /// Inserts the sheet at the given position.
     pub fn insert_sheet(&mut self, i: usize, sheet: Sheet) {
-        self.sheets.insert(i, sheet);
+        self.sheets.insert(i, sheet.into());
     }
 
     /// Appends a sheet.
     pub fn push_sheet(&mut self, sheet: Sheet) {
-        self.sheets.push(sheet);
+        self.sheets.push(sheet.into());
     }
 
     /// Removes a sheet from the table.
+    ///
+    /// Panics
+    ///
+    /// Panics if the sheet was detached.
     pub fn remove_sheet(&mut self, n: usize) -> Sheet {
-        self.sheets.remove(n)
+        self.sheets.remove(n).take()
     }
 
     /// Adds a default-style for all new values.
@@ -395,7 +477,10 @@ impl WorkBook {
     }
 
     /// Adds a style.
-    pub fn add_tablestyle(&mut self, style: TableStyle) -> TableStyleRef {
+    pub fn add_tablestyle(&mut self, mut style: TableStyle) -> TableStyleRef {
+        if style.name().is_empty() {
+            style.set_name(auto_style_name(&mut self.autonum, "ta", &self.tablestyles));
+        }
         let sref = style.style_ref();
         self.tablestyles.insert(style.name().to_string(), style);
         sref
@@ -417,7 +502,10 @@ impl WorkBook {
     }
 
     /// Adds a style.
-    pub fn add_rowstyle(&mut self, style: RowStyle) -> RowStyleRef {
+    pub fn add_rowstyle(&mut self, mut style: RowStyle) -> RowStyleRef {
+        if style.name().is_empty() {
+            style.set_name(auto_style_name(&mut self.autonum, "ro", &self.rowstyles));
+        }
         let sref = style.style_ref();
         self.rowstyles.insert(style.name().to_string(), style);
         sref
@@ -439,7 +527,10 @@ impl WorkBook {
     }
 
     /// Adds a style.
-    pub fn add_colstyle(&mut self, style: ColStyle) -> ColStyleRef {
+    pub fn add_colstyle(&mut self, mut style: ColStyle) -> ColStyleRef {
+        if style.name().is_empty() {
+            style.set_name(auto_style_name(&mut self.autonum, "co", &self.colstyles));
+        }
         let sref = style.style_ref();
         self.colstyles.insert(style.name().to_string(), style);
         sref
@@ -461,7 +552,10 @@ impl WorkBook {
     }
 
     /// Adds a style.
-    pub fn add_cellstyle(&mut self, style: CellStyle) -> CellStyleRef {
+    pub fn add_cellstyle(&mut self, mut style: CellStyle) -> CellStyleRef {
+        if style.name().is_empty() {
+            style.set_name(auto_style_name(&mut self.autonum, "ce", &self.cellstyles));
+        }
         let sref = style.style_ref();
         self.cellstyles.insert(style.name().to_string(), style);
         sref
@@ -483,7 +577,14 @@ impl WorkBook {
     }
 
     /// Adds a style.
-    pub fn add_paragraphstyle(&mut self, style: ParagraphStyle) -> ParagraphStyleRef {
+    pub fn add_paragraphstyle(&mut self, mut style: ParagraphStyle) -> ParagraphStyleRef {
+        if style.name().is_empty() {
+            style.set_name(auto_style_name(
+                &mut self.autonum,
+                "pa",
+                &self.paragraphstyles,
+            ));
+        }
         let sref = style.style_ref();
         self.paragraphstyles.insert(style.name().to_string(), style);
         sref
@@ -505,7 +606,10 @@ impl WorkBook {
     }
 
     /// Adds a style.
-    pub fn add_textstyle(&mut self, style: TextStyle) -> TextStyleRef {
+    pub fn add_textstyle(&mut self, mut style: TextStyle) -> TextStyleRef {
+        if style.name().is_empty() {
+            style.set_name(auto_style_name(&mut self.autonum, "te", &self.textstyles));
+        }
         let sref = style.style_ref();
         self.textstyles.insert(style.name().to_string(), style);
         sref
@@ -527,7 +631,14 @@ impl WorkBook {
     }
 
     /// Adds a style.
-    pub fn add_graphicstyle(&mut self, style: GraphicStyle) -> GraphicStyleRef {
+    pub fn add_graphicstyle(&mut self, mut style: GraphicStyle) -> GraphicStyleRef {
+        if style.name().is_empty() {
+            style.set_name(auto_style_name(
+                &mut self.autonum,
+                "gr",
+                &self.graphicstyles,
+            ));
+        }
         let sref = style.style_ref();
         self.graphicstyles.insert(style.name().to_string(), style);
         sref
@@ -615,6 +726,25 @@ impl WorkBook {
     }
 }
 
+#[derive(Clone, Debug)]
+pub struct WorkBookConfig {
+    pub active_table: String,
+    pub show_grid: bool,
+    pub show_page_breaks: bool,
+    pub has_sheet_tabs: bool,
+}
+
+impl Default for WorkBookConfig {
+    fn default() -> Self {
+        Self {
+            active_table: "".to_string(),
+            show_grid: true,
+            show_page_breaks: false,
+            has_sheet_tabs: true,
+        }
+    }
+}
+
 /// Visibility of a column or row.
 #[derive(Debug, Clone, Copy, Eq, PartialEq)]
 pub enum Visibility {
@@ -661,6 +791,8 @@ struct RowHeader {
     style: Option<String>,
     cellstyle: Option<String>,
     visible: Visibility,
+    repeat: u32,
+    height: Length,
 }
 
 impl RowHeader {
@@ -669,6 +801,8 @@ impl RowHeader {
             style: None,
             cellstyle: None,
             visible: Default::default(),
+            repeat: 1,
+            height: Default::default(),
         }
     }
 
@@ -703,6 +837,22 @@ impl RowHeader {
     pub fn visible(&self) -> Visibility {
         self.visible
     }
+
+    pub fn set_repeat(&mut self, repeat: u32) {
+        self.repeat = repeat;
+    }
+
+    pub fn repeat(&self) -> u32 {
+        self.repeat
+    }
+
+    pub fn set_height(&mut self, height: Length) {
+        self.height = height;
+    }
+
+    pub fn height(&self) -> Length {
+        self.height
+    }
 }
 
 /// Column data
@@ -711,6 +861,7 @@ struct ColHeader {
     style: Option<String>,
     cellstyle: Option<String>,
     visible: Visibility,
+    width: Length,
 }
 
 impl ColHeader {
@@ -719,6 +870,7 @@ impl ColHeader {
             style: None,
             cellstyle: None,
             visible: Default::default(),
+            width: Default::default(),
         }
     }
 
@@ -753,6 +905,14 @@ impl ColHeader {
     pub fn visible(&self) -> Visibility {
         self.visible
     }
+
+    pub fn set_width(&mut self, width: Length) {
+        self.width = width;
+    }
+
+    pub fn width(&self) -> Length {
+        self.width
+    }
 }
 
 /// One sheet of the spreadsheet.
@@ -777,6 +937,8 @@ pub struct Sheet {
     header_cols: Option<ColRange>,
     print_ranges: Option<Vec<CellRange>>,
 
+    sheet_config: SheetConfig,
+
     extra: Vec<XmlTag>,
 }
 
@@ -816,6 +978,7 @@ impl Sheet {
             header_rows: None,
             header_cols: None,
             print_ranges: None,
+            sheet_config: Default::default(),
             extra: vec![],
             row_header: Default::default(),
             display: true,
@@ -836,6 +999,7 @@ impl Sheet {
             header_rows: None,
             header_cols: None,
             print_ranges: None,
+            sheet_config: Default::default(),
             extra: Default::default(),
         }
     }
@@ -853,6 +1017,7 @@ impl Sheet {
             header_rows: self.header_rows.clone(),
             header_cols: self.header_cols.clone(),
             print_ranges: self.print_ranges.clone(),
+            sheet_config: Default::default(),
             extra: self.extra.clone(),
         }
     }
@@ -867,6 +1032,14 @@ impl Sheet {
         &self.name
     }
 
+    pub fn config(&self) -> &SheetConfig {
+        &self.sheet_config
+    }
+
+    pub fn config_mut(&mut self) -> &mut SheetConfig {
+        &mut self.sheet_config
+    }
+
     /// Sets the table-style
     pub fn set_style(&mut self, style: &TableStyleRef) {
         self.style = Some(style.to_string());
@@ -944,41 +1117,42 @@ impl Sheet {
         }
     }
 
-    /// Creates a col style and sets the col width.
-    pub fn set_col_width(&mut self, workbook: &mut WorkBook, col: ucell, width: Length) {
-        let style_name = format!("co{}", col);
+    /// Sets the column width for this column.
+    pub fn set_col_width(&mut self, col: ucell, width: Length) {
+        self.col_header
+            .entry(col)
+            .or_insert_with(ColHeader::new)
+            .set_width(width);
+    }
 
-        let mut colstyle = if let Some(style) = workbook.remove_colstyle(&style_name) {
-            style
+    /// Returns the column-width.
+    pub fn col_width(&self, col: ucell) -> Length {
+        if let Some(ch) = self.col_header.get(&col) {
+            ch.width()
         } else {
-            ColStyle::new(&style_name)
-        };
-        colstyle.set_col_width(width);
-        colstyle.set_use_optimal_col_width(false);
-        let colstyle = workbook.add_colstyle(colstyle);
-
-        self.set_colstyle(col, &colstyle);
+            Length::Default
+        }
     }
 
     /// Row style.
-    pub fn set_rowstyle(&mut self, col: ucell, style: &RowStyleRef) {
+    pub fn set_rowstyle(&mut self, row: ucell, style: &RowStyleRef) {
         self.row_header
-            .entry(col)
+            .entry(row)
             .or_insert_with(RowHeader::new)
             .set_style(style);
     }
 
     /// Remove the style.
-    pub fn clear_rowstyle(&mut self, col: ucell) {
+    pub fn clear_rowstyle(&mut self, row: ucell) {
         self.row_header
-            .entry(col)
+            .entry(row)
             .or_insert_with(RowHeader::new)
             .clear_style();
     }
 
     /// Returns the row style.
-    pub fn rowstyle(&self, col: ucell) -> Option<&String> {
-        if let Some(row_header) = self.row_header.get(&col) {
+    pub fn rowstyle(&self, row: ucell) -> Option<&String> {
+        if let Some(row_header) = self.row_header.get(&row) {
             row_header.style()
         } else {
             None
@@ -986,24 +1160,24 @@ impl Sheet {
     }
 
     /// Default cell style for this row.
-    pub fn set_row_cellstyle(&mut self, col: ucell, style: &CellStyleRef) {
+    pub fn set_row_cellstyle(&mut self, row: ucell, style: &CellStyleRef) {
         self.row_header
-            .entry(col)
+            .entry(row)
             .or_insert_with(RowHeader::new)
             .set_cellstyle(style);
     }
 
     /// Remove the style.
-    pub fn clear_row_cellstyle(&mut self, col: ucell) {
+    pub fn clear_row_cellstyle(&mut self, row: ucell) {
         self.row_header
-            .entry(col)
+            .entry(row)
             .or_insert_with(RowHeader::new)
             .clear_cellstyle();
     }
 
     /// Returns the default cell style for this row.
-    pub fn row_cellstyle(&self, col: ucell) -> Option<&String> {
-        if let Some(row_header) = self.row_header.get(&col) {
+    pub fn row_cellstyle(&self, row: ucell) -> Option<&String> {
+        if let Some(row_header) = self.row_header.get(&row) {
             row_header.cellstyle()
         } else {
             None
@@ -1011,36 +1185,57 @@ impl Sheet {
     }
 
     /// Visibility of the row
-    pub fn set_row_visible(&mut self, col: ucell, visible: Visibility) {
+    pub fn set_row_visible(&mut self, row: ucell, visible: Visibility) {
         self.row_header
-            .entry(col)
+            .entry(row)
             .or_insert_with(RowHeader::new)
             .set_visible(visible);
     }
 
     /// Returns the default cell style for this row.
-    pub fn row_visible(&self, col: ucell) -> Visibility {
-        if let Some(row_header) = self.row_header.get(&col) {
+    pub fn row_visible(&self, row: ucell) -> Visibility {
+        if let Some(row_header) = self.row_header.get(&row) {
             row_header.visible()
         } else {
             Default::default()
         }
     }
 
-    /// Creates a row-style and sets the row height.
-    pub fn set_row_height(&mut self, workbook: &mut WorkBook, row: ucell, height: Length) {
-        let style_name = format!("ro{}", row);
+    /// Sets the repeat count for this row. Usually this is the last row
+    /// with data in a sheet. Setting the repeat count will not change
+    /// the row number of following rows. But they will be changed after
+    /// writing to an ODS file and reading it again.
+    pub fn set_row_repeat(&mut self, row: ucell, repeat: u32) {
+        self.row_header
+            .entry(row)
+            .or_insert_with(RowHeader::new)
+            .set_repeat(repeat)
+    }
 
-        let mut rowstyle = if let Some(style) = workbook.remove_rowstyle(&style_name) {
-            style
+    /// Returns the repeat count for this row.
+    pub fn row_repeat(&self, row: ucell) -> u32 {
+        if let Some(row_header) = self.row_header.get(&row) {
+            row_header.repeat()
         } else {
-            RowStyle::new(&style_name)
-        };
-        rowstyle.set_row_height(height);
-        rowstyle.set_use_optimal_row_height(false);
-        let rowstyle = workbook.add_rowstyle(rowstyle);
+            Default::default()
+        }
+    }
 
-        self.set_rowstyle(row, &rowstyle);
+    /// Sets the row-height.
+    pub fn set_row_height(&mut self, row: ucell, height: Length) {
+        self.row_header
+            .entry(row)
+            .or_insert_with(RowHeader::new)
+            .set_height(height);
+    }
+
+    /// Returns the row-height
+    pub fn row_height(&self, row: ucell) -> Length {
+        if let Some(rh) = self.row_header.get(&row) {
+            rh.height()
+        } else {
+            Length::Default
+        }
     }
 
     /// Returns a tuple of (max(row)+1, max(col)+1)
@@ -1235,6 +1430,58 @@ impl Sheet {
     }
 }
 
+#[derive(Clone, Copy, Debug)]
+pub enum SheetSplitMode {
+    None = 0,
+    Pixel = 1,
+    Cell = 2,
+}
+
+impl TryFrom<i16> for SheetSplitMode {
+    type Error = OdsError;
+
+    fn try_from(n: i16) -> Result<Self, Self::Error> {
+        match n {
+            0 => Ok(SheetSplitMode::None),
+            1 => Ok(SheetSplitMode::Pixel),
+            2 => Ok(SheetSplitMode::Cell),
+            _ => Err(OdsError::Ods(format!("Invalid split mode {}", n))),
+        }
+    }
+}
+
+#[derive(Clone, Debug)]
+pub struct SheetConfig {
+    pub cursor_x: ucell,
+    pub cursor_y: ucell,
+    pub hor_split_mode: SheetSplitMode,
+    pub vert_split_mode: SheetSplitMode,
+    pub hor_split_pos: ucell,
+    pub vert_split_pos: ucell,
+    /// 0 - 4 indicates the quadrant where the focus should be.
+    pub active_split_range: i16,
+    pub zoom_type: i16,
+    pub zoom_value: i32,
+    pub show_grid: bool,
+}
+
+impl Default for SheetConfig {
+    fn default() -> Self {
+        Self {
+            cursor_x: 0,
+            cursor_y: 0,
+            hor_split_mode: SheetSplitMode::None,
+            vert_split_mode: SheetSplitMode::None,
+            hor_split_pos: 0,
+            vert_split_pos: 0,
+            active_split_range: 2,
+            zoom_type: 0,
+            zoom_value: 100,
+            show_grid: true,
+        }
+    }
+}
+
 /// One Cell of the spreadsheet.
 #[derive(Debug, Clone, Default)]
 pub struct SCell {
